"""parserc.py: Parser Combinators
"""
# Standard library imports

# Third party imports
from parsec import parsecmap, desc, regex, string, joint as __joint, generate, one_of, exclude, many, count

# Local application imports


#
# Definitions
#
# Helper functions
def flatten(iter):
    return ''.join(iter)


# Overloaded parsec functions
# def count(p, n):
#     return parsec.parsecmap(parsec.count(p, n), flatten)


# def many(p):
#     return parsec.parsecmap(parsec.many(p), flatten)


def joint(*p):
    return parsecmap(__joint(*p), flatten)


# Generic parsers
nibble = desc(regex(r'[0-9a-zA-Z]'), 'nibble')
byte = desc(regex(r'[0-9a-zA-Z]{2}'), 'byte')
null = desc(string('00'), 'null')



"""binary.py
"""
# Standard library imports
from __future__ import annotations
from collections import UserString
from collections.abc import Sequence
import math
from typing import Iterable, SupportsInt
from functools import reduce
from operator import __add__, __xor__, __and__, __or__, __lshift__, __rshift__
import sys as _sys
from array import array

# Third party imports

# Local application imports


# Helper functions
__HEX_DIGITS = '0123456789abcdefABCDEF'
__BIT_DIGITS = '01'


def _clean_char(keep: str, ignore: str, message: str):
    def _clean_char(c: str):
        if c in keep:
            return c
        elif c in ignore:
            return ''
        else:
            raise TypeError(F"Character '{c}' not allowed in {message}")

    return _clean_char


def _clean_string(s: str, keep: str, ignore: str, message: str) -> str:
    return reduce(__add__, map(_clean_char(keep, ignore, message), s))


def _clean_hex_string(s: str, ignore: str = "_") -> str:
    return _clean_string(s, keep=__HEX_DIGITS, ignore=ignore, message="hexadecimal string")


def _clean_bit_string(s: str, ignore: str = "_") -> str:
    return _clean_string(s, keep=__BIT_DIGITS, ignore=ignore, message="bit string")


def _align_string(left, right):
    match len(left)-len(right):
        case l if l < 0:
            # left is shorter
            return left.lpad(len(right)), right
        case l if l == 0:
            # both are same length
            return left, right
        case l if l > 0:
            # left is longer
            return left, right.lpad(len(left))


def _bitwise_operation(left, right, op):
    match left, right:
        case ByteString(), ByteString():
            fmt = '02X'
        case HexString(), HexString():
            fmt = 'X'
        case BitString(), BitString():
            fmt = 'X'

    return ''.join(([format(op(int(l), int(r)), fmt) for l, r in zip(*_align_string(left, right))]))


# 'BitString' class
class BitString(UserString):
    def __init__(self, value: int | Sequence, *, ignore: str = "_"):
        if isinstance(value, int):
            value_string = format(value, "b")
        elif isinstance(value, str):
            value_string = _clean_bit_string(value, ignore)
        elif isinstance(value, bytes):
            value_string = ''.join([format(b, '08b') for b in value])
        elif isinstance(value, list):
            value_string = ''.join([format(b, '08b') for b in value])
        else:
            raise TypeError(
                F"BitString()| unsupported initializer type: {type(value)}")

        super().__init__(value_string)

    def __int__(self) -> int:
        return int(self.data, 2)

    def __repr__(self):
        return F"BitString('{self}')"

    def __hash__(self):
        return hash(self.data)

    def __eq__(self, string: SupportsInt):
        return int(self) == int(string)

    def __lt__(self, string: SupportsInt):
        return int(self) < int(string)

    def __le__(self, string: SupportsInt):
        return int(self) <= int(string)

    def __gt__(self, string: SupportsInt):
        return int(self) > int(string)

    def __ge__(self, string: SupportsInt):
        return int(self) >= int(string)

    def __or__(self, other: BitString) -> BitString:
        return BitString(_bitwise_operation(self, other, __or__))

    def __xor__(self, other: BitString) -> BitString:
        return BitString(_bitwise_operation(self, other, __xor__))

    def __and__(self, other: BitString) -> BitString:
        return BitString(_bitwise_operation(self, other, __and__))

    def __invert__(self) -> BitString:
        return self ^ BitString('1' * len(self))

    def lpad(self, size: int) -> BitString:
        if len(self) < size:
            return BitString('0' * (size-len(self))) + self
        else:
            return self

    def rpad(self, size: int) -> BitString:
        if len(self) < size:
            return self + BitString('0' * (size-len(self)))
        else:
            return self

    def blocks(self, blocksize: int) -> Iterable[BitString]:
        nr_blocks = math.ceil(len(self) / blocksize)
        return (self[i*blocksize:(i+1)*blocksize] for i in range(nr_blocks))

    # def join(self, seq: Iterable[BitString]) -> BitString:
    #     return reduce(__add__, seq)

    def permute(self, permutation: list[int]) -> BitString:
        str_out: str = ''

        for i in permutation:
            str_out += str(self[i-1])

        return BitString(str_out)

    def expand(self, expansion: list[int]) -> BitString:
        return self.permute(expansion)

    def left_circular_shit(self, shift: int) -> BitString:
        return self[shift:] + self[:shift]

    @property
    def byte_string(self) -> ByteString:
        if len(self) % 8 == 0:
            return ByteString([int(b) for b in self.blocks(8)])
        else:
            return ByteString([int(b) for b in self.zfill(len(self)+8 - len(self) % 8).blocks(8)])


# 'HexString' class
class HexString(UserString):
    def __init__(self, value: int | Sequence, *, ignore: str = "_"):
        if isinstance(value, int):
            value_string = format(value, "X")
        elif isinstance(value, str):
            value_string = _clean_hex_string(value, ignore)
        elif isinstance(value, bytes):
            value_string = ''.join([format(b, '02X') for b in value])
        elif isinstance(value, list):
            value_string = ''.join([format(b, '02X') for b in value])
        else:
            raise TypeError(
                F"HexString()| unsupported initializer type: {type(value)}")

        super().__init__(value_string)

    def __int__(self) -> int:
        return int(self.data, 16)

    def __str__(self):
        return str(self.data.upper())

    def __repr__(self):
        return F"HexString('{self}')"

    def __hash__(self):
        return hash(self.data.upper())

    def __eq__(self, string: SupportsInt):
        return int(self) == int(string)

    def __lt__(self, string: SupportsInt):
        return int(self) < int(string)

    def __le__(self, string: SupportsInt):
        return int(self) <= int(string)

    def __gt__(self, string: SupportsInt):
        return int(self) > int(string)

    def __ge__(self, string: SupportsInt):
        return int(self) >= int(string)

    def __or__(self, other: HexString) -> HexString:
        return HexString(_bitwise_operation(self, other, __or__))

    def __xor__(self, other: HexString) -> HexString:
        return HexString(_bitwise_operation(self, other, __xor__))

    def __and__(self, other: HexString) -> HexString:
        return HexString(_bitwise_operation(self, other, __and__))

    def __invert__(self) -> HexString:
        return self ^ HexString('F' * len(self))

    def lpad(self, size: int) -> HexString:
        if len(self) < size:
            return HexString('0' * (size-len(self))) + self
        else:
            return self

    def rpad(self, size: int) -> HexString:
        if len(self) < size:
            return self + HexString('0' * (size-len(self)))
        else:
            return self

    # @property
    # def bit_length(self) -> int:
    #     return len(self) * 4

    @property
    def dscan_decimalize(self) -> str:
        """dscan_decimalize: double scan decimalization
        """
        dstr1 = ''
        dstr2 = ''
        for h in self.data:
            if h.isdigit():
                dstr1 = dstr1 + h
            else:
                dstr2 = dstr2 + F"{int(h, 16)-10}"

        return dstr1 + dstr2

    def blocks(self, blocksize: int) -> Iterable[HexString]:
        nr_blocks = math.ceil(len(self) / blocksize)
        return (self[i*2*blocksize:(i+1)*2*blocksize] for i in range(nr_blocks))

    # def join(self, seq: Iterable[HexString]) -> HexString:
    #     return reduce(__add__, seq)

    @property
    def bytestring(self) -> ByteString:
        if len(self) % 2 == 0:
            return ByteString(str(self))
        else:
            return ByteString('0' + str(self))


# 'ByteString' class
class ByteString(HexString):
    def __init__(self, value: int | Sequence, *, ignore: str = "_"):
        if isinstance(value, int):
            value_string = format(value, "X")
            if len(value_string) % 2 != 0:
                value_string = '0' + value_string
        elif isinstance(value, str):
            value_string = _clean_hex_string(value, ignore)
            if len(value_string) % 2 != 0:
                raise ValueError(
                    F"ByteString()| received off number of nibbles: {value_string}")
        elif isinstance(value, bytes):
            value_string = ''.join([format(b, '02X') for b in value])
        elif isinstance(value, list):
            value_string = ''.join([format(int(b), '02X') for b in value])
        else:
            raise TypeError(
                F"ByteString()| unsupported initializer type: {type(value)}")

        super().__init__(value_string)

    def __int__(self) -> int:
        return int(self.data, 16)

    def __str__(self):
        return str(self.data.upper())

    def __repr__(self):
        return F"ByteString('{self.data.upper()}')"

    def __getitem__(self, key) -> ByteString:
        return ByteString(self.bytes[key])

    def __len__(self) -> int:
        return len(self.bytes)

    def __hash__(self):
        return hash(self.data.upper())

    def __eq__(self, string: SupportsInt):
        if isinstance(string, str):
            return int(self) == int(string, 16)
        else:
            return int(self) == int(string)

    def __lt__(self, string: SupportsInt):
        return int(self) < int(string)

    def __le__(self, string: SupportsInt):
        return int(self) <= int(string)

    def __gt__(self, string: SupportsInt):
        return int(self) > int(string)

    def __ge__(self, string: SupportsInt):
        return int(self) >= int(string)

    def __or__(self, other: ByteString) -> ByteString:
        return ByteString(_bitwise_operation(self, other, __or__))

    def __xor__(self, other: ByteString) -> ByteString:
        return ByteString(_bitwise_operation(self, other, __xor__))

    def __and__(self, other: ByteString) -> ByteString:
        return ByteString(_bitwise_operation(self, other, __and__))

    def __invert__(self) -> ByteString:
        return self ^ ByteString('FF' * len(self))

    @property
    def bytes(self) -> bytes:
        return bytes.fromhex(self.data)

    @property
    def list(self):
        return [int(b) for b in self.bytes]

    @property
    def array(self):
        return array('B', self.bytes)

    @property
    def bit_string(self) -> BitString:
        return BitString(self.bytes)

    def blocks(self, blocksize: int) -> Iterable[ByteString]:
        nr_blocks = math.ceil(len(self) / blocksize)
        return (self[i*blocksize:(i+1)*blocksize] for i in range(nr_blocks))

    def lpad(self, size: int) -> ByteString:
        return self.zfill(2*size)

    def rpad(self, size: int) -> ByteString:
        if len(self) < size:
            return self + ByteString('00' * (size-len(self)))
        else:
            return self

    def mask(self, mask: int) -> ByteString:
        return self & ByteString(mask)

    def is_bit_set(self, bit: int) -> bool:
        return self.mask(1 << (bit-1)) == 1 << (bit-1)

    def is_bit_unset(self, bit: int) -> bool:
        return not self.is_bit_set(bit)

    # def __rshift__(self, shift: int) -> ByteString:
    #     return ByteString.from_int(self.int >> shift).lpad(len(self))

    def startswith(self, prefix, start=0, end=_sys.maxsize):
        if isinstance(prefix, str):
            return self.data.upper().startswith(prefix.upper(), start, end)
        else:
            return self.data.upper().startswith(map(upper, prefix), start, end)



# from .apdu import CommandCase, CommandField, ResponseProcessingState, CommandApdu, Lc, StatusBytes, ResponseApdu
# from .encodings import TransmissionProtocol, Chaining, SecureMessaging, CLA, LengthFieldNc, LengthFieldNe, Lc, Le
# from .commands import FileControlInformation, FileOccurrence, GetResponse, Select, Selection



# from .bulk_message import BulkMessage, BulkOutMessage, BulkInMessage, CommandStatus, IccStatus, MessageType, Protocol, SlotError
# from .bulk_out_messages import PC_to_RDR_Abort, PC_to_RDR_GetParameters, PC_to_RDR_GetSlotStatus, PC_to_RDR_IccPowerOff, PC_to_RDR_IccPowerOn, PC_to_RDR_XfrBlock, PowerSelection
# from .bulk_in_messages import ClockStatus, RDR_to_PC, RDR_to_PC_DataBlock, RDR_to_PC_NotifySlotChange, RDR_to_PC_Parameters, RDR_to_PC_SlotStatus, SlotChangedStatus, SlotCurrentState
# from .descriptors import DeviceDescriptor, ConfigurationDescriptor, InterfaceDescriptor, SmartCardDeviceDescriptor, EndpointDescriptor, EndpointDirection, EndpointTransfertType, parse_descriptors, CcidFeatures



"""descriptors.py
"""
# Standard library imports
from array import array
from typing import Union
from copy import deepcopy
from enum import IntEnum

# Third party imports

# Local application imports
from common.intlist import to_hstr


# Enum definitions
class DescriptorType(IntEnum):
    Device = 1
    Configuration = 2
    String = 3
    Interface = 4
    Endpoint = 5
    DeviceQualifier = 6
    OtherSpeedConfiguration = 7
    InterfacePower = 8
    SmartCardDevice = 0x21


class EndpointDirection(IntEnum):
    Out = 0x00
    In = 0x80


class EndpointTransfertType(IntEnum):
    Control = 0x00
    Isochronous = 0x01
    Bulk = 0x02
    Interrupt = 0x03


class CcidFeatures(IntEnum):
    NoSpecialCharacteristics = 0x00000000
    AutomaticParameterConfigurationBasedOnAtrData = 0x00000002
    AutomaticActivationOfIccOnInserting = 0x00000004
    AutomaticIccVoltageSelection = 0x00000008
    AutomaticIccClockFrequencyChangeAccordingToActiveParametersProvidedByTheHostOrSelfDetermined = 0x00000010
    AutomaticBaudRateChangeAccordingToActiveParametersProvidedByTheHostOrSelfDetermined = 0x00000020
    AutomaticParametersNegotiationMadeByTheCcid = 0x00000040
    AutomaticPpsMadeByTheCcidAccordingToTheActiveParameters = 0x00000080
    CcidCanSetIccInClockStopMode = 0x00000100
    NadValueOtherThan00AcceptedForT1 = 0x00000200
    AutomaticIfsdExchangeAsFirstExchangeForT1 = 0x00000400
    TpduLevelExchangeWithCcid = 0x00010000
    ShortApduLevelExchangesWithCcid = 0x00020000
    ShortAndExtendedApduLevelExchangesWithCcid = 0x00040000
    UsbWakeUpSignalingSupportedOnCardInsertionAndRemoval = 0x00100000


# Class definitions
class DeviceDescriptor:
    def __init__(self, descriptor: Union[array, list[int]]):
        if descriptor[1] != DescriptorType.Device.value:
            raise ValueError(
                F"Expecting Device descriptor type 01, received {descriptor[1]:02X} ({DescriptorType(descriptor[1]).name})")

        if isinstance(descriptor, array):
            self._descriptor = descriptor[0:descriptor[0]]
        elif isinstance(descriptor, list):
            self._descriptor = array('B', descriptor[0:descriptor[0]])
        else:
            raise TypeError(
                F"Message should be array or list[int], received {type(descriptor)}")

    @property
    def bDeviceClass(self):
        return self._descriptor[4]

    @property
    def idVendor(self):
        return self._descriptor[8:10]

    @property
    def idProduct(self):
        return self._descriptor[10:12]

    @property
    def bNumConfigurations(self):
        return self._descriptor[17]

    @property
    def vendor(self):
        return F"{int.from_bytes(self.idVendor, byteorder='little'):04X}"

    @property
    def product(self):
        return F"{int.from_bytes(self.idProduct, byteorder='little'):04X}"

    def __str__(self):
        return to_hstr(self._descriptor)


class ConfigurationDescriptor:
    def __init__(self, descriptor: Union[array, list[int]]):
        if descriptor[1] != DescriptorType.Configuration.value:
            raise ValueError(
                F"Expecting Configuration descriptor type 02, received {descriptor[1]:02X} ({DescriptorType(descriptor[1]).name})")

        if isinstance(descriptor, array):
            self._descriptor = descriptor[0:descriptor[0]]
        elif isinstance(descriptor, list):
            self._descriptor = array('B', descriptor[0:descriptor[0]])
        else:
            raise TypeError(
                F"Message should be array or list[int], received {type(descriptor)}")

    @property
    def wTotalLength(self):
        return self._descriptor[2:4]

    @property
    def bNumInterfaces(self):
        return self._descriptor[4]

    @property
    def bConfigurationValue(self):
        return self._descriptor[5]

    @property
    def total_length(self):
        return int.from_bytes(self.wTotalLength, byteorder='little')

    def __str__(self):
        return to_hstr(self._descriptor)


class InterfaceDescriptor:
    def __init__(self, descriptor: Union[array, list[int]]):
        if descriptor[1] != DescriptorType.Interface.value:
            raise ValueError(
                F"Expecting Interface descriptor type 02, received {descriptor[1]:02X} ({DescriptorType(descriptor[1]).name})")

        if isinstance(descriptor, array):
            self._descriptor = descriptor[0:descriptor[0]]
        elif isinstance(descriptor, list):
            self._descriptor = array('B', descriptor[0:descriptor[0]])
        else:
            raise TypeError(
                F"Message should be array or list[int], received {type(descriptor)}")

    @property
    def bInterfaceNumber(self):
        return self._descriptor[2]

    @property
    def bNumEndpoints(self):
        return self._descriptor[4]

    @property
    def bInterfaceClass(self):
        return self._descriptor[5]

    def __str__(self):
        return to_hstr(self._descriptor)


class EndpointDescriptor:
    def __init__(self, descriptor: Union[array, list[int]]):
        if descriptor[1] != DescriptorType.Endpoint.value:
            raise ValueError(
                F"Expecting Endpoint descriptor type 05, received {descriptor[1]:02X} ({DescriptorType(descriptor[1]).name})")

        if isinstance(descriptor, array):
            self._descriptor = descriptor[0:descriptor[0]]
        elif isinstance(descriptor, list):
            self._descriptor = array('B', descriptor[0:descriptor[0]])
        else:
            raise TypeError(
                F"Message should be array or list[int], received {type(descriptor)}")

    @property
    def bEndpointAddress(self):
        return self._descriptor[2]

    @property
    def bmAttributes(self):
        return self._descriptor[3]

    @property
    def wMaxPacketSize(self):
        return self._descriptor[4:6]

    @property
    def bInterval(self):
        return self._descriptor[6]

    @property
    def number(self):
        return self.bEndpointAddress & 0x0F

    @property
    def direction(self):
        return EndpointDirection(self.bEndpointAddress & 0x80)

    @property
    def transfert_type(self):
        return EndpointTransfertType(self.bmAttributes & 0x03)

    @property
    def maximum_packet_size(self):
        return int.from_bytes(self.wMaxPacketSize, byteorder='little') & 0x07FF

    def __str__(self):
        return to_hstr(self._descriptor)


class SmartCardDeviceDescriptor:
    def __init__(self, descriptor: Union[array, list[int]]):
        if descriptor[1] != DescriptorType.SmartCardDevice.value:
            raise ValueError(
                F"Expecting Smart Card Device descriptor type 21, received {descriptor[1]:02X} ({DescriptorType(descriptor[1]).name})")

        if descriptor[0] != 0x36:
            raise ValueError(
                F"Smart Card Device descriptor should be 54 bytes, received {descriptor[0]:02X} ({to_hstr(descriptor[0:descriptor[0]])})")

        if isinstance(descriptor, array):
            self._descriptor = descriptor[0:descriptor[0]]
        elif isinstance(descriptor, list):
            self._descriptor = array('B', descriptor[0:descriptor[0]])
        else:
            raise TypeError(
                F"Message should be array or list[int], received {type(descriptor)}")

    @property
    def bMaxSlotIndex(self):
        return self._descriptor[4]

    @property
    def bVoltageSupport(self):
        return self._descriptor[5]

    @property
    def dwProtocols(self):
        return self._descriptor[6:10]

    @property
    def dwMaxIFSD(self):
        return self._descriptor[28:32]

    @property
    def dwFeatures(self):
        return self._descriptor[40:44]

    @property
    def dwMaxCCIDMessageLength(self):
        return self._descriptor[44:48]

    @property
    def bMaxCCIDBusySlots(self):
        return self._descriptor[53]

    @property
    def features(self):
        _features = int.from_bytes(self.dwFeatures, byteorder='little')
        if _features == 0x00000000:
            return [CcidFeatures.NoSpecialCharacteristics]
        else:
            _features_list = []
            for feature in CcidFeatures:
                if (feature & _features) != 0:
                    _features_list.append(feature)

            return _features_list

    def __str__(self):
        return F"Smart Card Device Class Descriptor: {to_hstr(self._descriptor)}"



"""key_management.py
"""

# Standard library imports

# Third party imports

# Local application imports
from common.binary import ByteString
from crypto.des import tdea_2_ede_ecb, tdea_2_ede_cbc


# CARD KEYS
def S_ENC_DK(S_ENC_KMC: ByteString, derivation_id: ByteString):
    return tdea_2_ede_ecb(S_ENC_KMC, ByteString(F"{derivation_id}F001{derivation_id}0F01"))


def S_MAC_DK(S_MAC_KMC: ByteString, derivation_id: ByteString):
    return tdea_2_ede_ecb(S_MAC_KMC, ByteString(F"{derivation_id}F002{derivation_id}0F02"))


def DEK_DK(DEK_KMC: ByteString, derivation_id: ByteString):
    return tdea_2_ede_ecb(DEK_KMC, ByteString(F"{derivation_id}F003{derivation_id}0F03"))


# SESSION KEYS
def C_MAC_SK(S_MAC_DK: ByteString, sequence_counter: ByteString):
    return tdea_2_ede_cbc(S_MAC_DK, ByteString(F"0101{sequence_counter}" + "00" * 12), ByteString('00' * 8))


def R_MAC_SK(S_MAC_DK: ByteString, sequence_counter: ByteString):
    return tdea_2_ede_cbc(S_MAC_DK, ByteString(F"0102{sequence_counter}" + "00" * 12), ByteString('00' * 8))


def S_ENC_SK(S_ENC_DK: ByteString, sequence_counter: ByteString):
    return tdea_2_ede_cbc(S_ENC_DK, ByteString(F"0182{sequence_counter}" + "00" * 12), ByteString('00' * 8))


def DEK_SK(DEK_DK: ByteString, sequence_counter: ByteString):
    return tdea_2_ede_cbc(DEK_DK, ByteString(F"0181{sequence_counter}" + "00" * 12), ByteString('00' * 8))



"""ber.py
"""
# Standard library imports
from enum import IntEnum
from math import log2

# Third party imports


# Local application imports
from common.binary import HexString, ByteString
from common.parserc import null, byte, nibble, joint, one_of, exclude, generate, many, count


# Enum definitions
class TagClass(IntEnum):
    Universal = 0x00
    Application = 0x40
    ContextSpecific = 0x80
    Private = 0xC0


class TagConstruction(IntEnum):
    Primitive = 0x00
    Constructed = 0x20


#
# Class definitions
#
# Tag: tag identifier
class Tag(ByteString):
    def __init__(self, tag: str):
        super().__init__(tag)
        tag, remainder = T_fieldP.parse_partial(self.data)
        if remainder != '':
            raise ValueError(
                F"Unknown bytes '{remainder}' after tag '{tag}'")

        if len(self) == 1:
            if (tag_number := int(self) & 0x1F) == 0x1F:
                raise ValueError(
                    F"Tag(): 1-byte tag should not have b5-b1 = 11111, received: {self}")
            else:
                self.__tag_number = tag_number
                return

        if (int(self[0]) & 0x1F) != 0x1F:
            raise ValueError(
                F"Tag(): first byte of multi-byte tags should have b5-b1 = 11111, received: {self}")

        if len(self) > 2:
            if any(map(lambda n: (n & 0x80) == 0x00, self.list[1:-1])):
                raise ValueError(
                    F"Tag(): All but last tag byte should have b8 = 1, received: {self}")

        if (int(self[-1]) & 0x80) != 0x00:
            raise ValueError(
                F"Tag(): Last tag byte should have b8 = 0, received {self}")

        __subsequent_bytes = map(lambda n: n & 0x7F, self.list[2:])
        __tag_number = int(self[1]) & 0x7F
        for b in __subsequent_bytes:
            __tag_number = __tag_number << 7
            __tag_number += b

        self.__tag_number = __tag_number
        return

    @property
    def class_(self) -> TagClass:
        return TagClass(self.list[0] & 0xC0)

    @property
    def construction(self) -> TagConstruction:
        return TagConstruction(self.list[0] & 0x20)

    @property
    def number(self) -> int:
        return self.__tag_number


def create_tag(tag_class: TagClass, tag_construction: TagConstruction, tag_number: int) -> Tag:
    """create_tag(): create a Tag object based on individual fields
    """
    match tag_number:
        case number if number < 0:
            raise ValueError(
                F"create_tag(): Tag number should be positive, received: {tag_number}")

        case number if number < 31:
            tag = [tag_class+tag_construction+tag_number]

        case number if number < 128:
            tag = [tag_class+tag_construction+31, tag_number]

        case _:
            tag = [tag_class+tag_construction+31]

            subsequent_bytes = []
            while tag_number > 0:
                subsequent_bytes.append(128+(tag_number % 128))
                tag_number //= 128

            subsequent_bytes.reverse()
            subsequent_bytes[-1] &= 0x7F
            tag.extend(subsequent_bytes)

    return Tag(''.join([F"{b:02X}" for b in tag]))


# Length: length encoding
class Length(ByteString):
    def __init__(self, length: str):
        super().__init__(length)
        length, remainder = L_fieldP.parse_partial(self.data)
        if remainder != '':
            raise ValueError(
                F"Unknown bytes '{remainder}' after length '{length}'")

        length_bytes = self.bytes

        if len(length_bytes) == 1:
            if (length_bytes[0] & 0x80) == 0:
                self.__length = length_bytes[0]
            else:
                raise ValueError(
                    F"Single-byte length should have b8 = 0, received: {self}")
        elif len(length_bytes[1:]) == length_bytes[0] & 0x7F:
            self.__length = int.from_bytes(length_bytes[1:], byteorder='big')
        else:
            raise ValueError(
                F"Length should be {(length_bytes[0] & 0x7F) + 1} bytes, received: {self}")

    @property
    def value(self) -> int:
        return self.__length


def create_length(value: ByteString) -> Length:
    match len(value):
        case l if l < 128:
            return Length(F'{l:02X}')

        case l if log2(l) < 1017:
            nr_bits = len(value).bit_length()
            if nr_bits % 8 == 0:
                nr_bytes = nr_bits//8
            else:
                nr_bytes = nr_bits//8 + 1
            return Length(F"{128+nr_bytes:02X}" + format(len(value), F"0{nr_bytes*2}X"))

        case _:
            raise ValueError(F"Cannot encode length > 2^1016")


# TagLengthValue: TLV encoding
class TagLengthValue(ByteString):
    def __init__(self, tlv: str):
        super().__init__(tlv)
        partial_parse = TagLengthValueP.parse_partial(self.data)
        if partial_parse[1] != '':
            raise ValueError(
                F"Unknown bytes '{partial_parse[1]}' after TLV '{tlv}'")

        tag, length, value = partial_parse[0]
        self.__tag = Tag(tag)
        self.__length = Length(length)
        self.__value = ByteString(value)

    @property
    def tag(self) -> Tag:
        return self.__tag

    @property
    def length(self) -> Length:
        return self.__length

    @property
    def value(self) -> HexString:
        return self.__value


# class TagLength():
#     def __init__(self, tl: str):
#         pass


# class DataObjectList():
#     def __init__(self, dol: str):
#         pass


#
# Parser combinators
#
# Tag field parsers
__T_multi_byte_start = joint(one_of('13579bBdDfF'),
                             one_of('fF'))
__T_multi_byte_inner = joint(one_of('89aAbBcCdDeEfF'),
                             nibble)
__T_multi_byte_last = joint(one_of('01234567'),
                            nibble)
__T_single_byte = exclude(byte,
                          __T_multi_byte_start)


@generate
def __T_multi_byte():
    """Parses a multi-byte Tag Field.
    """
    start_tag = yield __T_multi_byte_start
    inner_tags = yield many(__T_multi_byte_inner)
    inner_tags = ''.join(inner_tags)
    last_tag = yield __T_multi_byte_last

    return start_tag + inner_tags + last_tag


T_fieldP = __T_single_byte | __T_multi_byte


# Length Field parser
__L_single_byte = joint(one_of('01234567'), nibble)
__L_multi_byte_start = joint(
    one_of('89aAbBcCdDeEfF'), nibble)


@generate
def __L_multi_byte():
    start_length = yield __L_multi_byte_start
    nr_bytes = int(start_length, 16) - 0x80
    subsequent_length = yield count(byte, nr_bytes)
    subsequent_length = ''.join(subsequent_length)

    return start_length + subsequent_length


L_fieldP = __L_single_byte | __L_multi_byte


# TLV parser
@generate
def TagLengthValueP():
    # ignore leading 00
    _ = yield many(null)

    tag = yield T_fieldP
    length = yield L_fieldP
    value = yield count(byte, Length(length).value)
    value = ''.join(value)

    # ignore trailing 00
    _ = yield many(null)

    return tag, length, value


# Old definitions
# def parse_to_dict(tlv_hstr: str):
#     pass


# def parse_dol(tlv_hstr: str):
#     pass

# def find(tag, tlv_object):
#     pass



"""scp02.py
"""

# Standard library imports
from enum import StrEnum

# Third party imports

# Local application imports
from common.binary import ByteString
from iso7816.apdu import CommandApdu, ResponseApdu
from crypto.des import tdea_2_ede_cbc, mac_2_ede


# Secure Channel Protocol '02' encodings defined by GP
class SecurityLevel(StrEnum):
    NoSecureMessagingExpected = '00'
    C_MAC = '01'
    C_DECRYPTIONandC_MAC = '03'
    R_MAC = '10'
    C_MACandR_MAC = '11'
    C_DECRYPTIONandC_MACandR_MAC = '13'


# MAC
def mac(C_MAC_SK: ByteString, data: ByteString, iv: ByteString = ByteString('00'*8)) -> ByteString:
    block = data + '80'
    if len(block) % 8 != 0:
        block += '00' * (8 - len(block) % 8)

    mac = mac_2_ede(C_MAC_SK, block)

    return mac


# Secure# CARD AUTHENTICATION CRYPTOGRAM
def card_challenge(C_MAC_SK: ByteString, AID: ByteString) -> ByteString:
    return mac(C_MAC_SK, AID)[0:6]
    # block = AID + '80'
    # if len(block) % 8 != 0:
    #     block += '00' * (8 - len(block) % 8)
    # mac = mac_2_ede(C_MAC_SK, block)
    # return mac[0:6]


def card_cryptogram(S_ENC_SK: ByteString, host_challenge: ByteString, sequence_counter: ByteString, card_challenge: ByteString) -> ByteString:
    card_cryptogram_input = host_challenge + \
        sequence_counter + card_challenge + '80' + '00' * 7
    return tdea_2_ede_cbc(S_ENC_SK, card_cryptogram_input, ByteString('00' * 8))[-8:]


# HOST AUTHENTICATION CRYPTOGRAM
def host_cryptogram(S_ENC_SK: ByteString, sequence_counter: ByteString, card_challenge: ByteString, host_challenge: ByteString) -> ByteString:
    host_cryptogram_input = sequence_counter + \
        card_challenge + host_challenge + '80' + '00' * 7
    return tdea_2_ede_cbc(S_ENC_SK, host_cryptogram_input, ByteString('00' * 8))[-8:]


# HOST MAC
def host_mac(C_MAC_SK: ByteString, host_cryptogram: ByteString) -> ByteString:
    return mac(C_MAC_SK, "8482010010" + host_cryptogram)
    # return mac_2_ede(C_MAC_SK, "8482010010" + host_cryptogram + '80' + '00' * 2)


#
# APDUs
#
# Secure Channel Protocol '02' command APDUs defined by GP
class InitializeUpdate(CommandApdu):
    def __init__(self, CLA: ByteString, key_version_number: ByteString, host_challenge: ByteString):
        if len(host_challenge) != 8:
            raise ValueError(
                F"INITIALIZE UPDATE: host cryptogram should be 8 bytes, received '{host_challenge}'")

        return super().__init__(CLA, ByteString('50'), key_version_number, ByteString('00'), data_field=host_challenge, Ne=256)


def INITIALIZE_UPDATE(CLA: ByteString, key_version_number: ByteString, host_challenge: ByteString):
    return InitializeUpdate(CLA, key_version_number, host_challenge)


class InitializeUpdateResponse(ResponseApdu):
    def __init__(self, response: ByteString):
        if len(response) != 30:
            raise ValueError(
                F"INITIALIZE UPDATE: response should be 30 bytes, received '{response}'")

        return super().__init__(response)

    @property
    def key_diversification_data(self):
        return self[0:10]

    @property
    def key_information(self):
        return self[10:12]

    @property
    def sequence_counter(self):
        return self[12:14]

    @property
    def card_challenge(self):
        return self[14:20]

    @property
    def card_cryptogram(self):
        return self[20:28]


class ExternalAuthenticate(CommandApdu):
    def __init__(self, CLA: ByteString, security_level: SecurityLevel, host_cryptogram: ByteString, MAC: ByteString):
        if len(host_cryptogram) != 8:
            raise ValueError(
                F"EXTERNAL AUTHENTICATE: host cryptogram should be 8 bytes, received '{host_cryptogram}'")

        if len(MAC) != 8:
            raise ValueError(
                F"EXTERNAL AUTHENTICATE: MAC should be 8 bytes, received '{MAC}'")

        return super().__init__(CLA, ByteString('82'), ByteString(security_level.value), ByteString('00'), data_field=host_cryptogram+MAC, Ne=None)


def EXTERNAL_AUTHENTICATE(CLA: ByteString, security_level: ByteString, host_cryptogram: ByteString, MAC: ByteString):
    return ExternalAuthenticate(CLA, SecurityLevel(security_level), host_cryptogram, MAC)



"""commands.py
"""

# Standard library imports
from typing import Optional


# Third party imports

# Local application imports
from common.binary import ByteString
from iso7816.apdu import CommandApdu
from iso7816.encodings import ClassByte as ISO_ClassByte
from .encodings import ClassByte, GetDataObject, FileOccurrence, ApplicationIdentifier


# Command APDUs defined by GP
class GetData(CommandApdu):
    def __init__(self, CLA: ClassByte, tag: ByteString, data: Optional[ByteString]):
        header = (CLA + ByteString('CA') + tag).blocks(1)
        if data is None:
            super().__init__(*header, data_field=None, Ne=256)
        else:
            super().__init__(*header, data_field=data, Ne=256)


def GET_DATA(CLA: ClassByte, tag: ByteString | GetDataObject, data: Optional[ByteString]) -> GetData:
    """GET_DATA: generate APDU for GET DATA command
    """
    if isinstance(tag, GetDataObject):
        if len(tag.value) == 2:
            P1 = ByteString('00')
            P2 = ByteString(tag.value)
        else:
            P1 = ByteString(tag.value[0:2])
            P2 = ByteString(tag.value[2:4])

    elif isinstance(tag, ByteString):
        P1 = tag[0:2]
        P2 = tag[2:4]

    else:
        raise TypeError(
            F"globalplatform.GET_DATA(): tag should be of type ByteString or GetDataObject, received: {type(tag)}")

    return GetData(CLA, P1+P2, data)


class Select(CommandApdu):
    def __init__(self, CLA: ISO_ClassByte, file_occurrence: FileOccurrence, aid: ByteString | ApplicationIdentifier):
        header = (
            CLA + ByteString(F'A404{file_occurrence:02X}')).blocks(1)

        match aid:
            case ApplicationIdentifier():
                if aid == ApplicationIdentifier.Default:
                    super().__init__(*header, data_field=None, Ne=256)
                else:
                    super().__init__(*header, data_field=ByteString(aid.value), Ne=256)
            case ByteString():
                super().__init__(*header, data_field=aid, Ne=256)
            case _:
                raise TypeError(
                    F"Select(): aid should be of type ByteString or ApplicationIdentifier, received: {type(aid)}")


def SELECT(CLA: ISO_ClassByte, file_occurrence: FileOccurrence, aid: ByteString | ApplicationIdentifier) -> Select:
    """SELECT(): generate CommandApdu for SELECT command (GlobalPlatform)
    """

    return Select(CLA, file_occurrence, aid)



"""commands.py
"""

# Standard library imports
from enum import StrEnum

# Third party imports

# Local application imports
# from common.ber import encode
from common.binary import ByteString
from iso7816.apdu import CommandApdu
from iso7816.encodings import Lc, Le


# Enum Definitions
class ApplicationIdentifier(StrEnum):
    Mastercard = 'A0000000041010'
    Visa = 'A0000000031010'
    Maestro = 'A00000000410103060'
    EVC = 'A0000008864040'
    EVC_test = 'A0000008863030'
    Dummy = 'A0000008862020'
    Dummy_test = 'A0000008861010'


class GetDataObject(StrEnum):
    ATC = '9F36'
    LastOnlineATCRegister = '9F13'
    PINTryCounter = '9F17'
    LogFormat = '9F4F'


# Command APDUs defined by EMV
class Select(CommandApdu):
    def __init__(self, aid: ByteString):
        header = ByteString('00A40400').blocks(1)
        super().__init__(*header,  data_field=aid, Ne=256)


def SELECT(aid: ByteString | ApplicationIdentifier) -> Select:
    """SELECT(): generate CommandApdu for SELECT command (EMV)
    """
    match aid:
        case ApplicationIdentifier():
            application_identifier = ByteString(aid.value)
        case ByteString():
            application_identifier = aid
        case _:
            raise TypeError(
                F"SELECT(): aid should be of type ByteString or ApplicationIdentifier, received: {type(aid)}")

    return Select(application_identifier)


class GetProcessingOptions(CommandApdu):
    def __init__(self, pdol: ByteString):
        header = ByteString('80A80000').blocks(1)
        super().__init__(*header, data_field=pdol, Ne=256)


def GET_PROCESSING_OPTIONS(pdol: ByteString | None) -> GetProcessingOptions:
    """GET_PROCESSING_OPTIONS(): generate APDU for GET PROCESSING OPTIONS command
    """
    if pdol:
        return GetProcessingOptions(pdol)
    else:
        return GetProcessingOptions(ByteString('8300'))


GPO = GET_PROCESSING_OPTIONS


class ReadRecord(CommandApdu):
    def __init__(self, SFI: int, record: int):
        header = ByteString(F"00B2{record:02X}{SFI*8+4}00").blocks(1)
        super().__init__(*header, data_field=None, Ne=None)


def READ_RECORD(SFI: int, record: int) -> ReadRecord:
    """READ_RECORD(): generate APDU for READ RECORD command
    """
    return ReadRecord(SFI, record)


class GetData(CommandApdu):
    def __init__(self, tag: ByteString):
        header = ByteString(F"80CA{tag}").blocks(1)
        super().__init__(*header, data_field=None, Ne=256)


def GET_DATA(tag: ByteString | GetDataObject) -> GetData:
    """GET_DATA: generate APDU for GET DATA command
    """
    if isinstance(tag, GetDataObject):
        if len(tag.value) == 2:
            P1 = ByteString('00')
            P2 = ByteString(tag.value)
        else:
            P1 = ByteString(tag.value[0:2])
            P2 = ByteString(tag.value[2:4])

    elif isinstance(tag, ByteString):
        P1 = tag[0:2]
        P2 = tag[2:4]

    else:
        raise TypeError(
            F"emv.GET_DATA(): tag should be of type ByteString or GetDataObject, received: {type(tag)}")

    return GetData(P1+P2)



"""encodings.py
"""

# Standard library imports
from enum import StrEnum, IntEnum

# Third party imports

# Local application imports
from common.binary import ByteString
from common.hstr import clean as _clean


# Enum Definitions
class SecureMessaging(IntEnum):
    No = 0x0
    GlobalPlatform = 0x4
    Iso7816 = 0x8
    Iso7816_CMAC = 0xC


class CardPersonalizationLifeCycleData(StrEnum):
    ICFabricator = "IC Fabricator"
    ICType = "IC Type"
    OperatingSystemIdentifier = "Operating System Identifier"
    OperatingSystemReleaseDate = "Operating System Release Date"
    OperatingSystemReleaseLevel = "Operating System Release Level"
    ICFabricationDate = "IC Fabrication Date"
    ICSerialNumber = "IC Serial Number"
    ICBatchIdentifier = "IC Batch Identifier"
    ICModuleFabricator = "IC Module Fabricator"
    ICModulePackagingDate = "IC Module Packaging Date"
    ICCManufacturer = "ICC Manufacturer"
    ICEmbeddingDate = "IC Embedding Date"
    ICPrePersonalizer = "IC Pre-Personalizer"
    ICPrePersonalizationDate = "IC Pre-Personalization Date"
    ICPrePersonalizationEquipmentIdentifier = "IC Pre-Personalization Equipment Identifier"
    ICPersonalizer = "IC Personalizer"
    ICPersonalizationDate = "IC Personalization Date"
    ICPersonalizationEquipmentIdentifier = "IC Personalization Equipment Identifier"


class GetDataObject(StrEnum):
    ListOfApplications = '2F00'
    IssuerIdentificationNumber = '0042'
    CardImageNumber = '0045'
    CardData = '0066'
    KeyInformationTemplate = '00E0'
    CardCapabilityInformation = '0067'
    CurrentSecurityLevel = '00D3'
    SecurityDomainManagerURL = '5F50'
    ConfirmationCounter = '00C2'
    SequenceCounterOfTheDefaultKeyVersionNumber = '00C1'
    CardProductionLifeCycle = '9F7F'


class FileOccurrence(IntEnum):
    FirstOrOnlyOccurrence = 0x00
    NextOccurrence = 0x02


class ApplicationIdentifier(StrEnum):
    GlobalPlatformSecurityDomain = 'A000000151000000'
    Default = ''


# Class Definitions
class ClassByte(ByteString):
    def __init__(self, secure_messaging: SecureMessaging, logical_channel: int):
        match logical_channel:
            case logical_channel if logical_channel < 0:
                raise ValueError(
                    F'Logical channel number out of bound: should be in [0;19], received {logical_channel}')

            case logical_channel if logical_channel <= 3:
                # CLA Byte Coding according to Table 11-11
                _class_byte = 0x80
                _class_byte += secure_messaging
                _class_byte += logical_channel

            case logical_channel if logical_channel <= 19:
                # CLA Byte Coding according to Table 11-12
                _class_byte = 0xC0 if secure_messaging == SecureMessaging.No else 0xE0
                _class_byte += 4 + logical_channel

            case _:
                raise ValueError(
                    F'Logical channel number out of bound: should be in [0;19], received {logical_channel}')

        super().__init__(_class_byte)

        self.__secure_messaging = secure_messaging
        self.__logical_channel = logical_channel

    @property
    def secure_messaging(self) -> SecureMessaging:
        return self.__secure_messaging

    @property
    def logical_channel(self) -> int:
        return self.__logical_channel


def CLA(value: int | str | ByteString) -> ClassByte:
    """CLA(): creates a GlobalPlatform-compliant CLA Byte
    """
    match value:
        case int():
            class_byte = ByteString(value)
        case str():
            class_byte = ByteString(value)
        case ByteString():
            class_byte = value

        case _:
            raise TypeError(
                F"CLA()| type {type(value)} not supported for argument value")

    if len(class_byte) != 1:
        raise ValueError(
            F"CLA()| expecting 1 byte, received {class_byte} bytes")

    if class_byte.is_bit_unset(8):
        # GP CLA Byte must have b8 set
        raise ValueError(F"CLA(): expecting b8 = 1, received {class_byte}")

    if class_byte.is_bit_unset(7):
        # CLA Byte Coding according to Table 11-11
        secure_messaging = SecureMessaging(int(class_byte.mask(0x0C)))
        logical_channel = int(class_byte.mask(0x03))
    else:
        # CLA Byte Coding according to Table 11-12
        secure_messaging = SecureMessaging.No if class_byte.is_bit_unset(
            6) else SecureMessaging.GlobalPlatform
        logical_channel = 4 + int(class_byte.mask(0x0F))

    return ClassByte(secure_messaging, logical_channel)


class CPLC:
    def __init__(self, data: str):
        cplc_data = _clean(
            data, "globalplatform.encodings.CPLC.__init__()", "data")
        if len(cplc_data) != 84:
            raise ValueError(
                F"CPLC data should be 42 bytes, received {len(cplc_data)//2} bytes: {cplc_data}")
        else:
            self.__cplc_data = cplc_data

    def field(self, data: CardPersonalizationLifeCycleData):
        pos, length = _CPLC_DATA_DEFINITION[data]
        return self.__cplc_data[2*pos:2*(pos+length)]

    def dict(self):
        cplc_dict = {}
        for field in CardPersonalizationLifeCycleData:
            cplc_dict[field] = self.field(field)

        return cplc_dict

    def __str__(self):
        return self.__cplc_data


# Helper functions
_CPLC_DATA_DEFINITION = {
    CardPersonalizationLifeCycleData.ICFabricator: (0, 2),
    CardPersonalizationLifeCycleData.ICType: (2, 2),
    CardPersonalizationLifeCycleData.OperatingSystemIdentifier: (4, 2),
    CardPersonalizationLifeCycleData.OperatingSystemReleaseDate: (6, 2),
    CardPersonalizationLifeCycleData.OperatingSystemReleaseLevel: (8, 2),
    CardPersonalizationLifeCycleData.ICFabricationDate: (10, 2),
    CardPersonalizationLifeCycleData.ICSerialNumber: (12, 4),
    CardPersonalizationLifeCycleData.ICBatchIdentifier: (16, 2),
    CardPersonalizationLifeCycleData.ICModuleFabricator: (18, 2),
    CardPersonalizationLifeCycleData.ICModulePackagingDate: (20, 2),
    CardPersonalizationLifeCycleData.ICCManufacturer: (22, 2),
    CardPersonalizationLifeCycleData.ICEmbeddingDate: (24, 2),
    CardPersonalizationLifeCycleData.ICPrePersonalizer: (26, 2),
    CardPersonalizationLifeCycleData.ICPrePersonalizationDate: (28, 2),
    CardPersonalizationLifeCycleData.ICPrePersonalizationEquipmentIdentifier: (30, 4),
    CardPersonalizationLifeCycleData.ICPersonalizer: (34, 2),
    CardPersonalizationLifeCycleData.ICPersonalizationDate: (36, 2),
    CardPersonalizationLifeCycleData.ICPersonalizationEquipmentIdentifier: (38, 4),
}



"""test_common_ber.py
"""
# Standard library imports
import unittest

# Third party imports

# Local application imports
from common.ber import HexString, TagClass, TagConstruction, Tag, create_tag, Length, create_length, T_fieldP, L_fieldP, parserc, TagLengthValueP


#
# Test values
#


#
# Unit tests
#
class TestMethods(unittest.TestCase):
    def test_Tag(self):
        self.assertEqual(Tag('01').class_, TagClass.Universal)
        self.assertEqual(Tag('01').construction, TagConstruction.Primitive)
        self.assertEqual(Tag('01').number, 1)
        self.assertEqual(Tag('ED').class_, TagClass.Private)
        self.assertEqual(Tag('ED').construction, TagConstruction.Constructed)
        self.assertEqual(Tag('ED').number, 13)
        self.assertEqual(Tag('5F28').class_, TagClass.Application)
        self.assertEqual(Tag('5F28').construction, TagConstruction.Primitive)
        self.assertEqual(Tag('5F28').number, 40)

        with self.assertRaises(ValueError):
            Tag('5A080123456789ABCDEF')

    def test_create_tag(self):
        self.assertEqual(create_tag(TagClass.Universal, TagConstruction.Primitive, 1),
                         Tag('01'))
        self.assertEqual(create_tag(TagClass.Private, TagConstruction.Constructed, 13),
                         Tag('ED'))
        self.assertEqual(create_tag(TagClass.Application, TagConstruction.Primitive, 40),
                         Tag('5F28'))

    def test_Length(self):
        self.assertEqual(Length('05').value, 5)
        self.assertEqual(Length('8203FC').value, 1020)

        with self.assertRaises(ValueError):
            Length('5A080123456789ABCDEF')

    def test_create_length(self):
        self.assertEqual(create_length(HexString('FF'*5)), Length('05'))
        self.assertEqual(create_length(HexString('FF'*1020)), Length('8203FC'))

    def test_T_fieldP(self):
        self.assertEqual(T_fieldP.parse_partial('6F108408A000000003000000A5049F6501FF'),
                         ('6F', '108408A000000003000000A5049F6501FF'))
        self.assertEqual(T_fieldP.parse_partial('9F272D' + 'FF'*0x2d),
                         ('9F27', '2D' + 'FF'*0x2d))

    def test_L_fieldP(self):
        self.assertEqual(L_fieldP.parse_partial('108408A000000003000000A5049F6501FF'),
                         ('10', '8408A000000003000000A5049F6501FF'))
        self.assertEqual(L_fieldP.parse_partial('8180FF'),
                         ('8180', 'FF'))

    def test_TagLengthValueP(self):
        self.assertEqual(TagLengthValueP.parse_partial('8408A000000003000000A5049F6501FF'),
                         (('84', '08', 'A000000003000000'), 'A5049F6501FF'))
        self.assertEqual(parserc.many(TagLengthValueP).parse('8408A000000003000000A5049F6501FF'),
                         [('84', '08', 'A000000003000000'), ('A5', '04', '9F6501FF')])

    # def test_parse(self):
    #     self.assertEqual(find('6F', [('6F', '10', [('84', '08', 'A000000003000000'), ('A5', '04', [('9F65', '01', 'FF')])])]),
    #                      ('6F', '10', [('84', '08', 'A000000003000000'), ('A5', '04', [('9F65', '01', 'FF')])]))
    #     self.assertEqual(find('84', [('6F', '10', [('84', '08', 'A000000003000000'), ('A5', '04', [('9F65', '01', 'FF')])])]),
    #                      ('84', '08', 'A000000003000000'))
    #     self.assertEqual(find('A5', [('6F', '10', [('84', '08', 'A000000003000000'), ('A5', '04', [('9F65', '01', 'FF')])])]),
    #                      ('A5', '04', [('9F65', '01', 'FF')]))
    #     self.assertEqual(find('9F65', [('6F', '10', [('84', '08', 'A000000003000000'), ('A5', '04', [('9F65', '01', 'FF')])])]),
    #                      ('9F65', '01', 'FF'))

    # def test_parse_dol(self):
    #     self.assertEqual(parse_dol('9F1A02'), ([('9F1A', '02')], ''))


if __name__ == '__main__':
    unittest.main()



"""key_management.py
"""

# Standard library imports

# Third party imports
try:
    from icecream import ic
    ic.disable()
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

# Local application imports
from common import hstr
from common.binary import ByteString
from crypto import des


# Function definitions
def master_key_derivation_A(imk: ByteString, pan: str, psn: str = '00'):
    """master_key_derivation_A(): ICC Master Key derivation (Option A)
    """
    X = pan + psn
    if len(X) < 16:
        X = '0' * (16 - len(X)) + X

    Y = ByteString(X[-16:])
    ic(X, Y)

    Zl = des.adjust_parity(des.tdea_2_ede(imk, Y))
    Zr = des.adjust_parity(des.tdea_2_ede(imk, ~Y))
    ic(Zl, Zr)

    return Zl + Zr



# Standard library imports
from enum import IntEnum
from typing import Union

# Third party imports

# Local application imports
from .bulk_message import *


# Definitions
class SlotCurrentState(IntEnum):
    NoIccPresent = 0
    IccPresent = 1


class SlotChangedStatus(IntEnum):
    NoChange = 0
    Change = 1


class ClockStatus(IntEnum):
    ClockRunning = 0
    ClockStoppedStateL = 1
    ClockStoppedStateH = 2
    ClockStoppedUnknownState = 3


class RDR_to_PC_NotifySlotChange(BulkInMessage):
    def current_state(self, slot_nr: int) -> SlotCurrentState:
        slot_current_state = (self.bmSlotICCState >> (2 * slot_nr)) & 0x1
        return SlotCurrentState(slot_current_state)

    def changed_status(self, slot_nr: int) -> SlotCurrentState:
        slot_changed_status = (self.bmSlotICCState >>
                               (2 * slot_nr + 1)) & 0x1
        return SlotCurrentState(slot_changed_status)

    @property
    def bmSlotICCState(self):
        return self._message[1]

    def __str__(self):
        return F"{super().__str__()}, {self.current_state().name}, {self.changed_status().name}"


class RDR_to_PC_DataBlock(BulkInMessage):
    def __str__(self):
        if self.command_status == CommandStatus.Failed:
            return F"{super().__str__()}"
        else:
            return F"{super().__str__()}, data={self.data}"

    @property
    def bChainParameter(self):
        return self._message[9]


class RDR_to_PC_SlotStatus(BulkInMessage):
    @property
    def clock_status(self):
        return ClockStatus(self.bClockStatus)

    @property
    def bClockStatus(self):
        return self._message[9]

    def __str__(self):
        return F"{super().__str__()}, {self.clock_status.name}"


class RDR_to_PC_Parameters(BulkInMessage):
    @property
    def protocol(self):
        return Protocol(self.bProtocolNum)

    @property
    def bProtocolNum(self):
        return self._message[9]

    def __str__(self):
        return F"{super().__str__()}, protocol {self.protocol}"


# Functions
def RDR_to_PC(message: Union[array, list[int]]):
    cls = {0x50: RDR_to_PC_NotifySlotChange,
           0x80: RDR_to_PC_DataBlock,
           0x81: RDR_to_PC_SlotStatus,
           0x82: RDR_to_PC_Parameters}.get(message[0], BulkInMessage)

    return cls(message)



"""apdu.py
"""

# Standard library imports
from __future__ import annotations
from enum import StrEnum
from typing import Optional

# Third party imports

# Local application imports
from common.binary import ByteString
from .encodings import Lc, Le


# Enum Definitions
class CommandCase(StrEnum):
    Case1 = 'Case 1'
    Case2S = 'Case 2S'
    Case2E = 'Case 2E'
    Case3S = 'Case 3S'
    Case3E = 'Case 3E'
    Case4S = 'Case 4S'
    Case4E = 'Case 4E'


class CommandField(StrEnum):
    Header = 'Header'
    Body = 'Body'
    Class = 'Class'
    Instruction = 'Instruction'
    P1 = 'P1'
    P2 = 'P2'
    Lc = 'Lc'
    Data = 'Data'
    Le = 'Le'


class ResponseProcessingState(StrEnum):
    Normal = 'Normal procesing'
    Warning = 'Warning processing'
    ExecutionError = 'Execution error'
    CheckingError = 'Checking error'


# Command and Response APDU classes
class CommandApdu(ByteString):
    def __init__(self, CLA: ByteString, INS: ByteString, P1: ByteString, P2: ByteString, data_field: Optional[ByteString], Ne: Optional[int]):
        for param in ["CLA", "INS", "P1", "P2"]:
            if len(locals()[param]) != 1:
                raise ValueError(
                    F"{param} should be 1 byte, received: {locals()[param]}")

        match data_field, Ne:
            case None, None:
                # Case 1
                super().__init__(str(CLA + INS + P1 + P2))
                self.__case = CommandCase.Case1

            case None, int():
                # Case 2
                if len(Le(Ne)) == 1:
                    # Case 2S
                    super().__init__(str(CLA + INS + P1 + P2 + Le(Ne)))
                    self.__case = CommandCase.Case2S
                else:
                    # Case 2E
                    super().__init__(str(CLA + INS + P1 + P2 + Le(Ne)))
                    self.__case = CommandCase.Case2E

            case ByteString(), None:
                # Case 3
                if len(Lc(data_field)) == 1:
                    # Case 3S
                    super().__init__(str(CLA + INS + P1 + P2 + Lc(data_field) + data_field))
                    self.__case = CommandCase.Case3S
                else:
                    # Case 3E
                    super().__init__(str(CLA + INS + P1 + P2 + Lc(data_field) + data_field))
                    self.__case = CommandCase.Case3E

            case ByteString(), int():
                # Case 4
                if len(Le(Ne)) == 1:
                    # Case 4S
                    super().__init__(str(CLA + INS + P1 + P2 + Lc(data_field) + data_field + Le(Ne)))
                    self.__case = CommandCase.Case4S
                else:
                    # Case 4E
                    super().__init__(str(CLA + INS + P1 + P2 + Lc(data_field) + data_field + Le(Ne)))
                    self.__case = CommandCase.Case4E

            case _, _:
                raise ValueError(
                    F"Received wrong data_field ({data_field}) and/or Ne ({Ne})")

    @property
    def case(self) -> CommandCase:
        return self.__case

    @property
    def header(self) -> ByteString:
        return self[0:4]

    @property
    def body(self) -> ByteString:
        return self[4:]

    @property
    def CLA(self) -> ByteString:
        return self[0]

    @property
    def INS(self) -> ByteString:
        return self[1]

    @property
    def P1(self) -> ByteString:
        return self[2]

    @property
    def P2(self) -> ByteString:
        return self[3]

    @property
    def Lc(self) -> ByteString:
        match self.__case:
            case CommandCase.Case1 | CommandCase.Case2S | CommandCase.Case2E:
                raise ValueError(
                    F'Command APDU has no Lc field ({self.case.value})')

            case CommandCase.Case3S | CommandCase.Case4S:
                return self[4]

            case CommandCase.Case3E | CommandCase.Case4E:
                return self[4:7]

    @property
    def data_field(self) -> ByteString:
        match self.case:
            case CommandCase.Case1 | CommandCase.Case2S | CommandCase.Case2E:
                raise ValueError(
                    F'Command APDU has no Data field ({self.case.value})')

            case CommandCase.Case3S:
                return self[5:]

            case CommandCase.Case4S:
                return self[5:-1]

            case CommandCase.Case3E:
                return self[7:]

            case CommandCase.Case4E:
                return self[7:-2]

    @property
    def Le(self) -> ByteString:
        match self.case:
            case CommandCase.Case1 | CommandCase.Case3S | CommandCase.Case3E:
                raise ValueError(
                    F'Command APDU has no Le field ({self.case.value})')

            case CommandCase.Case2S | CommandCase.Case4S:
                return self[-1]

            case CommandCase.Case2E | CommandCase.Case4E:
                return self[-2:]

    @property
    def Ne(self) -> int:
        match self.case:
            case CommandCase.Case1 | CommandCase.Case3S | CommandCase.Case3E:
                raise ValueError(
                    F'Command APDU has no Le field ({self.case.value})')

            case CommandCase.Case2S | CommandCase.Case4S:
                return 256 if self[-1] == '00' else int(self[-1])

            case CommandCase.Case2E | CommandCase.Case4E:
                return 65536 if self[-2:] == '0000' else int(self[-2:])

    def updated_Ne(self, Ne: int) -> CommandApdu:
        match self.case:
            case CommandCase.Case1 | CommandCase.Case3S | CommandCase.Case3E:
                raise ValueError(
                    F'Command APDU has no Le field ({self.case.value})')

            case CommandCase.Case2S:
                return CommandApdu(self.CLA, self.INS, self.P1, self.P2, data_field=None, Ne=Ne)

            case CommandCase.Case4S:
                return CommandApdu(self.CLA, self.INS, self.P1, self.P2, data_field=self.data_field, Ne=Ne)

            case CommandCase.Case2E | CommandCase.Case4E:
                raise ValueError(
                    F"Cannot update Le with APDU {self.case.value}")


def CAPDU(apdu: str | ByteString) -> CommandApdu:
    """CAPDU(): creates a CommandApdu object
    """
    match apdu:
        case str():
            _apdu = ByteString(apdu)
        case ByteString():
            _apdu = apdu
        case _:
            raise TypeError(
                F"CAPDU(): type {type(apdu)} not supported for argument apdu")

    header = _apdu[0:4].blocks(1)
    match len(_apdu):
        case l if l < 4:
            raise ValueError(
                F'CAPDU(): wrong Command APDU length, expected 4 or more bytes but received {_apdu}')

        case 4:
            # Case 1
            return CommandApdu(*header, data_field=None, Ne=None)

        case 5:
            # Case 2S
            Ne = 256 if _apdu[4] == '00' else int(_apdu[4])
            return CommandApdu(*header, data_field=None, Ne=Ne)

        case _:
            if _apdu[4] != '00':
                if len(_apdu) == 5 + int(_apdu[4]):
                    # Case 3S
                    return CommandApdu(*header, data_field=_apdu[5:], Ne=None)

                elif len(_apdu) == 5 + int(_apdu[4]) + 1:
                    # Case 4S
                    Ne = 256 if _apdu[4] == '00' else int(_apdu[4])
                    return CommandApdu(*header, data_field=_apdu[5:], Ne=Ne)

                else:
                    raise ValueError(
                        F'APDU(): wrong length, expected {5+int(_apdu[4])} bytes for Case 3s or {5+int(_apdu[4])+1} bytes for Case 4s but received {len(_apdu)} bytes instead')

            else:
                # Extended length cases
                if len(_apdu) == 7:
                    # Case 2E
                    Ne = 65536 if _apdu[4] == '00' else int(_apdu[4])
                    return CommandApdu(*header, data_field=None, Ne=Ne)

                elif len(_apdu) == 7 + int(_apdu[5:7]):
                    # Case 3E
                    return CommandApdu(*header, data_field=_apdu[7:], Ne=None)

                elif len(_apdu) == 7 + int(_apdu[5:7]) + 2:
                    # Case 4E
                    Ne = 65536 if _apdu[-2:] == '0000' else int(_apdu[-2:])
                    return CommandApdu(*header, data_field=_apdu[7:], Ne=None)

                else:
                    raise ValueError(
                        F'RAPDU(): wrong Command APDU length: expected {7 + int(_apdu[5:7])} bytes for Case 3e or {7 + int(_apdu[5:7]) + 2} bytes for Case 4e but received {len(_apdu)} bytes instead')


class StatusBytes(ByteString):
    def __init__(self, sw12: ByteString):
        if len(sw12) != 2:
            raise ValueError(F"Expected 2 bytes, received: {sw12}")

        if not (sw12.startswith('6') or sw12.startswith('9')):
            raise ValueError(F"Expected 6XXX or 9XXX value, received: {sw12}")

        if sw12.startswith('60'):
            raise ValueError(F"60XX in invalid, received: {sw12}")

        super().__init__(str(sw12))

    @property
    def state(self) -> ResponseProcessingState:
        match self:
            case '9000':
                return ResponseProcessingState.Normal
            case sw12 if sw12.startswith('61'):
                return ResponseProcessingState.Normal

            case sw12 if sw12.startswith('62') or sw12.startswith('63'):
                return ResponseProcessingState.Warning

            case sw12 if sw12.startswith('64') or sw12.startswith('65') or sw12.startswith('66'):
                return ResponseProcessingState.ExecutionError

            case sw12 if sw12.startswith('67') or sw12.startswith('68') or sw12.startswith('69') or sw12.startswith('6A') or sw12.startswith('6B') or sw12.startswith('6C') or sw12.startswith('6D') or sw12.startswith('6E') or sw12.startswith('6F'):
                return ResponseProcessingState.CheckingError

            case _:
                raise ValueError(F"Unknown processing state: {self}")

    @property
    def meaning(self) -> str:
        match self:
            case '9000':
                return 'No further qualification'

            case sw12 if sw12.startswith('61'):
                return 'SW2 encodes the number of data bytes still available'

            case sw12 if sw12.startswith('62'):
                match sw12[1]:
                    case '00':
                        return 'No information given'
                    case '81':
                        return 'Part of returned data may be corrupted'
                    case '82':
                        return 'End of file or record reached before reading Ne bytes'
                    case '83':
                        return 'Selected file deactivated'
                    case '84':
                        return 'File control information not formatted according to ISO7816-4 5.3.3'
                    case '85':
                        return 'Selected file in termination state'
                    case '86':
                        return 'No input data available from a sensor on the card'
                    case _:
                        return 'State of non-volatile memory is unchanged'

            case sw12 if sw12.startswith('63'):
                match sw12[1]:
                    case '00':
                        return 'No information given'
                    case '81':
                        return 'File filled up by the last write'
                    case sw2 if sw2.startswith('C'):
                        return F'Counter = {int(str(sw2), 16)-12}'
                    case _:
                        return 'State of non-volatile memory has changed(further qualification in SW2)'

            case sw12 if sw12.startswith('64'):
                match sw12[1]:
                    case '00':
                        return 'Execution error'
                    case '01':
                        return 'Immediate response required by the card'
                    case sw2 if int(str(sw2), 16) >= 0x02 and int(str(sw2), 16) <= 0x80:
                        return 'Triggering by the card'
                    case _:
                        return 'Unknown meaning'

            case sw12 if sw12.startswith('65'):
                match sw12[1]:
                    case '00':
                        return 'No information given'
                    case '81':
                        return 'Memory failure'
                    case _:
                        return 'Unknown meaning'

            case sw12 if sw12.startswith('66'):
                return 'Security-related issues'

            case '6700':
                return 'Wrong length; no further indication'

            case sw12 if sw12.startswith('68'):
                match sw12[1]:
                    case '00':
                        return 'No information given'
                    case '81':
                        return 'Logical channel not supported'
                    case '82':
                        return 'Secure messaging not supported'
                    case '83':
                        return 'Last command of the chain expected'
                    case '84':
                        return 'Command chaining not supported'
                    case _:
                        return 'Functions in CLA not supported (further qualification in SW2)'

            case sw12 if sw12.startswith('69'):
                match sw12[1]:
                    case '00':
                        return 'No information given'
                    case '81':
                        return 'Command incompatible with file structure'
                    case '82':
                        return 'Security status not satisfied'
                    case '83':
                        return 'Authentication method blocked'
                    case '84':
                        return 'Reference data not usable'
                    case '85':
                        return 'Conditions of use not satisfied'
                    case '86':
                        return 'Command not allowed (no current EF)'
                    case '87':
                        return 'Expected secure messaging data objects missing'
                    case '88':
                        return 'Incorrect secure messaging data objects'
                    case _:
                        return 'Command not allowed (further qualification in SW2)'

            case sw12 if sw12.startswith('6A'):
                match sw12[1]:
                    case '00':
                        return 'No information given'
                    case '80':
                        return 'Incorrect parameters in the command data field'
                    case '81':
                        return 'Function not supported'
                    case '82':
                        return 'File or application not found'
                    case '83':
                        return 'Record not found'
                    case '84':
                        return 'Not enough memory space in the file'
                    case '85':
                        return 'Nc inconsistent with TLV structure'
                    case '86':
                        return 'Incorrect parameters P1-P2'
                    case '87':
                        return 'Nc inconsistent with parameters P1-P2'
                    case '88':
                        return 'Referenced data or reference data not found (exact meaning depending on the command)'
                    case '89':
                        return 'File already exists'
                    case '8A':
                        return 'DF name already exists'
                    case _:
                        return 'Wrong parameters P1-P2 (further qualification in SW2)'

            case '6B00':
                return 'Wrong parameters P1-P2'

            case sw12 if sw12.startswith('6C'):
                return 'Wrong Le field; SW2 encodes the exact number of available data bytes'

            case '6D00':
                return 'Instruction code not supported or invalid'

            case '6E00':
                return 'Class not supported'

            case '6F00':
                return 'No precise diagnosis'

            case _:
                return 'Unkown meaning'


class ResponseApdu(ByteString):
    def __init__(self, response: ByteString):
        if len(response) < 2:
            raise ValueError(
                F"Expecting reponse of at least 2 bytes, received: {response}")
        super().__init__(str(response))

    @property
    def body(self) -> ByteString:
        if len(self) > 2:
            return self[0:-2]
        else:
            raise ValueError(F"No response body")

    @property
    def SW12(self) -> StatusBytes:
        return StatusBytes(self[-2:])

    @property
    def SW1(self) -> ByteString:
        return self.SW12[0]

    @property
    def SW2(self) -> ByteString:
        return self.SW12[1]

    @property
    def StatusBytes(self) -> StatusBytes:
        return self.SW12


def RAPDU(apdu: str | ByteString) -> ResponseApdu:
    """RAPDU(): creates a CommandApdu object
    """
    match apdu:
        case str():
            return ResponseApdu(ByteString(apdu))
        case ByteString():
            return ResponseApdu(apdu)
        case _:
            raise TypeError(
                F"RAPDU(): type {type(apdu)} not supported for argument apdu")



"""encodings.py
"""

# Standard library imports
from enum import StrEnum, IntEnum

# Third party imports

# Local application imports
from common.binary import ByteString


# Enum Definitions
class TransmissionProtocol(StrEnum):
    T0 = 'T=0'
    T1 = 'T=1'


class SecureMessaging(IntEnum):
    No = 0x0
    Proprietary = 0x4
    Iso7816 = 0x8
    Iso7816_CMAC = 0xC


class Chaining(StrEnum):
    LastOrOnly = 'The command is the last or only command of a chain'
    NotLast = 'The command is not the last command of a chain'


class Selection(IntEnum):
    SelectMForDForEF = 0x00
    SelectChildDF = 0x01
    SelectEFUnderCurrentDF = 0x02
    SelectParentDFOfCurrentDF = 0x03
    SelectByDFName = 0x04
    SelectFromMF = 0x08
    SelectFromCurrentDF = 0x09


class FileOccurrence(IntEnum):
    FirstOrOnlyOccurrence = 0x0
    LastOccurrence = 0x1
    NextOccurrence = 0x2
    PreviousOccurrence = 0x3


class FileControlInformation(IntEnum):
    ReturnFCITemplate = 0x00
    ReturnFCPTempalte = 0x10
    ReturnFMDTemplate = 0x20
    NoResponseOrProprietary = 0x30


# Class definitions
class ClassByte(ByteString):
    def __init__(self, secure_messaging: SecureMessaging, logical_channel: int, command_chaining: Chaining):
        match logical_channel:
            case logical_channel if logical_channel < 0:
                raise ValueError(
                    F'Logical channel number out of bound: should be in [0;19], received {logical_channel}')

            case logical_channel if logical_channel <= 3:
                # CLA Byte Coding according to Table 2
                _class_byte = secure_messaging
                _class_byte += logical_channel

            case logical_channel if logical_channel <= 19:
                # CLA Byte Coding according to Table 3
                _class_byte = 0x40 if secure_messaging == SecureMessaging.No else 0x60
                _class_byte += 4 + logical_channel

            case _:
                raise ValueError(
                    F'Logical channel number out of bound: should be in [0;19], received {logical_channel}')

        if command_chaining == Chaining.NotLast:
            _class_byte += 0x10

        super().__init__(_class_byte)

        self.__secure_messaging = secure_messaging
        self.__logical_channel = logical_channel
        self.__command_chaining = command_chaining

    @property
    def secure_messaging(self) -> SecureMessaging:
        return self.__secure_messaging

    @property
    def logical_channel(self) -> int:
        return self.__logical_channel

    @property
    def chaining(self) -> Chaining:
        return self.__command_chaining


def CLA(value: int | str | ByteString) -> ClassByte:
    """CLA(): creates an ISO7816-compliant CLA Byte
    """
    match value:
        case int():
            class_byte = ByteString(value)
        case str():
            class_byte = ByteString(value)
        case ByteString():
            class_byte = value

        case _:
            raise TypeError(
                F"CLA()| type {type(value)} not supported for argument value")

    if len(class_byte) != 1:
        raise ValueError(
            F"CLA()| expecting 1 byte, received {class_byte} bytes")

    if class_byte.is_bit_set(8):
        # ISO7816 CLA Byte must not have b8 set
        raise ValueError(F"CLA(): expecting b8 = 0, received {class_byte}")

    if class_byte.is_bit_unset(7):
        # CLA Byte Coding according to Table 2
        secure_messaging = SecureMessaging(int(class_byte.mask(0x0C)))
        logical_channel = int(class_byte.mask(0x03))
    else:
        # CLA Byte Coding according to Table 3
        secure_messaging = SecureMessaging.Iso7816 if class_byte.is_bit_set(
            6) else SecureMessaging.No
        logical_channel = 4 + int(class_byte.mask(0x0F))

    command_chaining = Chaining.LastOrOnly if class_byte.is_bit_unset(
        5) else Chaining.NotLast

    return ClassByte(secure_messaging, logical_channel, command_chaining)


class LengthFieldNc(ByteString):
    def __init__(self, Nc: int):
        match Nc:
            case n if n < 0:
                raise ValueError(
                    F"LengthFieldNc(): Ne must be > 0, received {Nc}")

            case 0:
                raise ValueError(F"LengthFieldNc(): Nc cannot be 0")

            case n if n < 256:
                super().__init__(F"{Nc:02X}")

            case n if n < 65536:
                super().__init__(F"00{Nc:04X}")

            case _:
                raise ValueError(
                    F"LengthFieldNc(): Nc must be < 65536, received {Nc}")


class LengthFieldNe(ByteString):
    def __init__(self, Ne: int):
        match Ne:
            case n if n < 0:
                raise ValueError(
                    F"LengthFieldNe(): Ne must be > 0, received {Ne}")

            case 0:
                raise ValueError(F"LengthFieldNe(): Ne cannot be 0")

            case n if n < 256:
                super().__init__(F"{Ne:02X}")

            case 256:
                super().__init__("00")

            case n if n < 65536:
                super().__init__(F"00{Ne:04X}")

            case 65536:
                super().__init__("000000")

            case _:
                raise ValueError(
                    F"LengthFieldNe(): Ne must be <= 65536, received {Ne}")


# Functions
def Lc(data_field: ByteString) -> LengthFieldNc:
    return LengthFieldNc(len(data_field))


Le = LengthFieldNe



"""commands.py
"""

# Standard library imports
from enum import Enum, IntEnum
from typing import Union

# Third party imports

# Local application imports
from common.hstr import clean as _clean
from multos.encodings import MultosData, MultosDataField
from iso7816 import CommandApdu, CommandField, CLA as ISO_CLA, Chaining


# Enum Definitions


# Command APDUs defined by MULTOS
class GetMultosData(CommandApdu):
    def __init__(self):
        super().__init__("800000007F")


def GET_MULTOS_DATA() -> GetMultosData:
    """GET_MULTOS_DATA: generate APDU for GET MULTOS DATA command
    """
    return GetMultosData()


class GetManufacturerData(CommandApdu):
    def __init__(self):
        super().__init__("8002000016")



"""commands.py
"""

# Standard library imports
from typing import Optional

# Third party imports

# Local application imports
from common.binary import ByteString
from .apdu import CommandApdu
from .encodings import CLA, Selection, FileOccurrence, FileControlInformation


#
# Commands for interchange
#
class GetResponse(CommandApdu):
    def __init__(self, class_byte: ByteString, Ne: int):
        header = (CLA(class_byte) + ByteString('C00000')).blocks(1)
        super().__init__(*header, data_field=None, Ne=Ne)


GET_RESPONSE = GetResponse


class Select(CommandApdu):
    def __init__(self,
                 class_byte: ByteString,
                 P1: ByteString,
                 P2: ByteString,
                 data_field: Optional[ByteString],
                 Ne: Optional[int]):
        if len(P1) != 1:
            raise ValueError(F"P1 should be 1 byte, received: {P1}")
        if len(P2) != 1:
            raise ValueError(F"P2 should be 1 byte, received: {P2}")

        header = (CLA(class_byte) + ByteString('A4') + P1 + P2).blocks(1)

        super().__init__(*header, data_field=data_field, Ne=Ne)


def SELECT(class_byte: ByteString,
           selection: Selection,
           file_occurrence: FileOccurrence,
           fci: FileControlInformation,
           data_field: Optional[ByteString],
           Ne: Optional[int]) -> Select:
    P1 = ByteString(F"{selection.value:02X}")
    P2 = ByteString(F"{file_occurrence.value + fci.value:02X}")

    return Select(class_byte, P1, P2, data_field, Ne)



"""evc_cvc3.py
"""
# Standard library imports

# Third party imports
try:
    from icecream import ic
    ic.disable()
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


# Local application imports
from common.binary import ByteString
from emv.key_management import master_key_derivation_A
from emv.dsc import generate_ivcvc3, generate_cvc3


# Example script
if __name__ == '__main__':
    ic.enable()
    ic.includeContext = True

    ic('Mastercard card data:')
    PAN = '5413123456784808'
    PSN = '00'
    EXPY_DATE = '1220'
    ic(PAN, PSN, EXPY_DATE)

    ic('Mastercard key derivation:')
    MDK = ByteString('01234567899876543210012345678998')
    ic(MDK)
    UDK = master_key_derivation_A(MDK, PAN, PSN)
    ic(UDK)

    ic('Mastercard IVCVC3 computation:')
    TRACK2_EQUIVALENT = F'{PAN}D{EXPY_DATE}0000000000000000F'
    ic(TRACK2_EQUIVALENT)
    IC_CVC3 = generate_ivcvc3(UDK, TRACK2_EQUIVALENT)
    ic(IC_CVC3)

    ic('Mastercard CVC3 computation')
    UNPREDICTABLE_NUMBER = '00000000'
    ic(UNPREDICTABLE_NUMBER)
    for ATC in [F'{atc:04X}' for atc in range(1, 6)]:
        block = F'{IC_CVC3}{UNPREDICTABLE_NUMBER}{ATC}'
        ic(ATC, block)
        CVC3 = generate_cvc3(UDK, block)[-3:]
        ic(CVC3)
